# Copyright (c) Qotto, 2021

from __future__ import annotations

import logging
import socket
import threading
from concurrent.futures import Future
from datetime import datetime
from threading import Thread, Lock
from time import sleep
from typing import Iterable, Optional
from uuid import uuid4

from confluent_kafka.admin import ClusterMetadata, TopicMetadata

from confluent_kafka import Producer, Consumer, Message, TopicPartition
from eventy.config import SERVICE_NAME
from eventy.messaging.errors import MessagingError
from eventy.messaging.store import RecordStoreWithInitialization
from eventy.record import Record
from eventy.serialization import RecordSerializer

__all__ = [
    'CKStore',
]

from eventy.trace_id import local_trace

logger = logging.getLogger(__name__)


def _message_str(message: Message) -> str:
    """
    String representation of kafka message formatted as "topic[partition]:offset".
    """
    return f'{message.topic()}[{message.partition()}]:{message.offset()}'


def _partitions_str(partitions: list[TopicPartition]) -> str:
    """
    String representation of a topic-partition list, including offsets and errors.
    """
    partitions_dict = {}
    for partition in partitions:
        if partition.topic not in partitions_dict:
            partitions_dict[partition.topic] = [partition.partition]
        else:
            partitions_dict[partition.topic].append(partition.partition)
    partitions_summary_str = ', '.join([f'{topic}{parts}' for topic, parts in partitions_dict.items()])
    partitions_complete_str = ', '.join(f'{p.topic}[{p.partition}]:{p.offset} (err:{p.error})' for p in partitions)
    return f'{partitions_summary_str}\n{partitions_complete_str}'


def _on_assign(consumer: Consumer, partitions: list[TopicPartition]) -> None:
    """
    Callback on partition assignment, logs the assigned partitions.
    """
    partitions_str = _partitions_str(partitions)
    logger.info(f"Kafka Subscribe: ASSIGNED partitions for consumer {consumer}: " + partitions_str)


def _on_revoke(consumer: Consumer, partitions: list[TopicPartition]) -> None:
    """
    Callback on partition revocation, logs the revoked partitions.
    """
    partitions_str = _partitions_str(partitions)
    logger.info(f"Kafka Subscribe: REVOKED partitions for consumer {consumer}: " + partitions_str)


def _on_lost(consumer: Consumer, partitions: list[TopicPartition]) -> None:
    """
    Callback on partition loss, logs the lost partitions.
    """
    partitions_str = _partitions_str(partitions)
    logger.info(f"Kafka Subscribe: LOST partitions for consumer {consumer}: " + partitions_str)


class CKStore(RecordStoreWithInitialization):
    """
    Kafka implementation of a record store, using Confluent kafka library
    """

    transactional_producer: Producer
    """ Kafka transactional producer, according to transactional state """
    immediate_producer: Producer
    """ Kafka immediate producer, always produce immediately """
    consumer: Consumer
    """ Kafka consumer, according to transactional state """
    poll_thread: Thread
    """ Thread for polling for messages for both producers"""

    def __init__(
        self,
        serializer: RecordSerializer,
        bootstrap_servers: list[str],
        group_id: Optional[str] = None,
        transactional_id: Optional[str] = None,
        sasl_username: Optional[str] = None,
        sasl_password: Optional[str] = None,
    ) -> None:
        """
        Create a new Kafka record store, the store will be initialized on first use, or with a direct call to initialize().

        :param RecordSerializer serializer: Record serializer (read and write)
        :param list[str] bootstrap_servers: Kafka bootstrap servers (producers and consumer)
        :param str group_id: Kafka group id (consumer), default: service name
        :param str transactional_id: Kafka transactional id for transactional producer, default is generate unique id
        :param Optional[str] sasl_username: Optional username (if SASL PLAIN kafka connection, producer and consumer)
        :param Optional[str] sasl_password: Optional password (if SASL PLAIN kafka connection, producer and consumer)
        """
        # Initialize topics to register
        super().__init__()
        # Kafka config
        self.serializer = serializer
        self.bootstrap_servers = bootstrap_servers
        self.group_id = group_id or SERVICE_NAME
        self.transactional_id = transactional_id or f'{SERVICE_NAME}-{str(uuid4())}'
        self.sasl_username = sasl_username
        self.sasl_password = sasl_password
        # Other attributes
        self.initialized = False
        """ Whether the store has been initialized """
        self.in_transaction = False
        """ Whether the store is in a transactional state """
        self.cancelled = False
        """ Whether the store has been cancelled """
        self._initialization_lock = Lock()
        """ Lock for initialization """

    def initialize(self) -> None:
        """
        Initialize the store if not already done.

        The store will stop polling for messages when the thread calling this method stops.
        """
        self._check_initialized()

    def _initialize_consumer_subscription(self):
        consumer_topics = [topic for topic in self.topics]
        logger.info(f"Kafka Subscribe: Initializing consumer subscription for topics {consumer_topics}")
        self.consumer.subscribe(
            consumer_topics,
            on_assign=_on_assign,
            on_revoke=_on_revoke,
            on_lost=_on_lost,
        )
        logger.debug(f"Initialized Kafka consumer.")

    def _initialize_consumer_assignment(self):
        consumer_topics = [topic for topic in self.assignments]
        logger.info(f"Kafka Assign: Initializing consumer assignment for topics {consumer_topics}: {self.assignments}")

        ck_topic_partitions: list[TopicPartition] = []
        ck_cluster_meta: ClusterMetadata = self.consumer.list_topics()

        for topic_name in consumer_topics:
            if topic_name not in ck_cluster_meta.topics:
                logger.warning(f"Topic {topic_name} not found.")
                continue
            else:
                logger.info(f"Topic {topic_name} found.")

            topic_meta: TopicMetadata = ck_cluster_meta.topics[topic_name]
            topic_offsets: Optional[list[int]] = self.assignments[topic_name][1]
            topic_since_date: Optional[datetime] = self.assignments[topic_name][0]

            if topic_offsets and len(topic_meta.partitions) != len(topic_offsets):
                logger.warning(
                    f"Topic {topic_name} has {len(topic_meta.partitions)} partitions, but {len(topic_offsets)} offsets."
                )
                continue

            for partition_id in topic_meta.partitions:
                ck_topic_partition: TopicPartition = TopicPartition(topic_name, partition_id)

                if topic_offsets:
                    ck_topic_partition.offset = topic_offsets[partition_id]
                    logger.info(f"Will assign {topic_name}[{partition_id}]:{ck_topic_partition.offset}")
                else:
                    ck_topic_partition.offset = int(topic_since_date.timestamp() * 1000)
                    ck_topic_partition = self.consumer.offsets_for_times([ck_topic_partition])[0]
                    logger.info(
                        f"Will assign {topic_name}[{partition_id}]:{ck_topic_partition.offset} "
                        f"(from date {topic_since_date})"
                    )
                ck_topic_partitions.append(ck_topic_partition)
            logger.info(f"Assigning {ck_topic_partitions}.")
            self.consumer.assign(ck_topic_partitions)
            logger.info(f"Assigned.")

    def _check_initialized(self) -> None:
        with self._initialization_lock:
            if self.initialized:
                logger.debug("Kafka store already initialized.")
                return

            logger.info(f"Will initialize Kafka producer and consumer.")

            # configs
            transactional_producer_config = {
                'bootstrap.servers': ','.join(self.bootstrap_servers),
                'transactional.id': self.transactional_id,
                'client.id': socket.gethostname(),
            }
            immediate_producer_config = {
                'bootstrap.servers': ','.join(self.bootstrap_servers),
                'client.id': socket.gethostname(),
                'enable.idempotence': 'true',
            }
            consumer_config = {
                'bootstrap.servers': ','.join(self.bootstrap_servers),
                'group.id': self.group_id,
                'client.id': socket.gethostname(),
                'enable.auto.commit': 'false',
                'auto.offset.reset': 'earliest',
                'isolation.level': 'read_committed',
            }

            # SASL authentication
            if self.sasl_username and self.sasl_password:
                for config in [
                    transactional_producer_config,
                    immediate_producer_config,
                    consumer_config,
                ]:
                    config.update(
                        {
                            'sasl_mechanism': 'PLAIN',
                            'sasl_plain_username': self.sasl_username,
                            'sasl_plain_password': self.sasl_password,
                        }
                    )

            # producers
            logger.info(f"Initializing transactional Kafka producer with config {transactional_producer_config}.")
            self.transactional_producer = Producer(transactional_producer_config)
            self.transactional_producer.init_transactions()
            logger.debug(f"Initialized transactional Kafka producer.")

            logger.info(f"Initializing immediate Kafka producer with config {immediate_producer_config}.")
            self.immediate_producer = Producer(immediate_producer_config)
            logger.debug(f"Initialized immediate Kafka producer.")

            # consumer
            logger.info(f"Initializing Kafka consumer with config {consumer_config}.")
            self.consumer = Consumer(consumer_config)
            if self.topics:
                self._initialize_consumer_subscription()
            elif self.assignments:
                self._initialize_consumer_assignment()
            else:
                logger.info(f"No subscription or assignments to initialize consumer with.")
            logger.debug(f"Initialized Kafka consumer.")

            # poll loop
            current_thread = threading.current_thread()

            def poll_loop():
                """
                Kafka requires that producers poll to actually send messages, and trigger callbacks.
                """

                while True:
                    if self.cancelled:
                        # the store was explicitly cancelled
                        finish_reason = "cancelled"
                        break
                    elif not current_thread.is_alive():
                        # the thread which called the initializer is not alive
                        finish_reason = "parent thread not alive"
                        break
                    else:
                        try:
                            immediate_polled = self.immediate_producer.poll(0.005)
                            if immediate_polled:
                                logger.debug(f"Immediate producer polled {immediate_polled} messages.")
                            transactional_polled = self.transactional_producer.poll(0.005)
                            if transactional_polled:
                                logger.debug(f"Transactional producer polled {transactional_polled} messages.")
                        except Exception as e:
                            # Exception could be temporary (e.g. kafka server not available)
                            logger.warning(f"Kafka exception in poll loop: {e}")
                            sleep(30)
                logger.warning(f"Kafka poll loop stopped. Reason: {finish_reason}.")

            self.poll_thread = Thread(
                name=f"CK-Poll-{uuid4().hex[-6:]}",
                target=poll_loop,
            )

            logger.debug(f"Will start Kafka poll loop thread.")
            self.poll_thread.start()

            self.initialized = True
            logger.debug(f"Record Store initialized.")

    def read(
        self,
        max_count: int = 1, timeout_ms: Optional[int] = None, auto_ack: bool = False
    ) -> Iterable[Record]:
        """
        Reads records from kafka, and returns them as a generator.

        :raise MessagingError: if there is an error reading from kafka.

        :param max_count: maximum number of records to fetch
        :param timeout_ms: maximum time to wait for records, in milliseconds
        :param auto_ack: whether to automatically acknowledge the records

        :return: generator of records
        """
        self._check_initialized()

        if not self.topics and not self.assignments:
            logger.debug(f"No topics configured. Returning empty list.")
            return []

        try:
            partitions_str = _partitions_str(self.consumer.assignment())
            logger.debug(f"Reading at most {max_count} records from Kafka. Current assignment is {partitions_str}.")
            messages: list[Message] = self.consumer.consume(max_count, _ms2sec_ck(timeout_ms))
            logger.debug(f"Read {len(messages)} messages from Kafka.")
        except Exception:
            # TODO: temporary errors? (are they always timeout?)
            raise MessagingError(f"Error consuming messages from kafka.")

        for message in messages:
            if message.error():
                logger.error(
                    f"Message {_message_str(message)} has error {message.error()}."
                )
                continue
            logger.debug(f"New message {_message_str(message)}.")
            try:
                # noinspection PyArgumentList
                # PyCharm is wrong, there is no payload arg in .value() method
                encoded = message.value()
                record = self.serializer.decode(encoded)
            except Exception as e:
                raise MessagingError(f"Error decoding message {_message_str(message)}.") from e
            with local_trace(correlation_id=record.correlation_id):
                # TODO: request_id could be Record.uuid, or Record.partition_key, or message offset
                logger.info(
                    f"New record {record.qualified_name} from kafka message {_message_str(message)}."
                )
                yield record
            if auto_ack:
                try:
                    self.consumer.commit(message=message, asynchronous=False)
                except Exception:
                    raise MessagingError(
                        f"Error committing message {_message_str(message)}."
                    )

    def has_reached_end_of_topics(self) -> bool:
        """
        Returns True if all topics have been consumed.
        """
        ck_topic_partitions = self.consumer.assignment()
        for ck_topic_partition in ck_topic_partitions:
            watermark = self.consumer.get_watermark_offsets(ck_topic_partition)
            position = self.consumer.position([ck_topic_partition])[0]
            if not watermark or not position:
                logger.warning(f"Could not get watermark or position for {ck_topic_partition}.")
                return False
            assigned_offset = ck_topic_partition.offset
            current_offset = position.offset
            current_watermark = watermark[1]
            if current_offset < 0:
                if assigned_offset < 0:
                    # There was no message in the partition when the consumer was created.
                    continue
            if current_offset < current_watermark:
                # There are still messages to be consumed.
                return False
        return True

    def ack(self, timeout_ms=None) -> None:
        """
        Acknowledges the last message read from kafka.

        If in transaction, the offsets are sent to transaction. Otherwise, the consumer is committed.
        """
        self._check_initialized()

        if self.in_transaction:
            logger.info(f"Ack read messages in transaction (will send offsets to transaction).")
            try:
                self.transactional_producer.send_offsets_to_transaction(
                    self.consumer.position(self.consumer.assignment()),
                    self.consumer.consumer_group_metadata(),
                    _ms2sec_ck(timeout_ms)
                )
            except Exception:
                raise MessagingError(f"Error sending offsets to transaction.")
        else:
            logger.info(f"Ack read messages outside transaction.")
            try:
                self.consumer.commit(asynchronous=False)
            except Exception as e:
                raise MessagingError(f"Error committing message.") from e

    def _write(self, producer: Producer, record: Record, topic: str, timeout_ms=None) -> None:

        # handle callback
        future_result: Future = Future()

        def ack(err, msg):
            with local_trace(correlation_id=record.correlation_id):
                if err:
                    logger.debug(f"Produced with errors: {err}.")
                    future_result.set_exception(MessagingError(err))
                else:
                    logger.debug(f"Produced successfully.")
                    future_result.set_result(msg)

        # produce message
        logger.debug(f"Will produce record {record.qualified_name} on {topic}.")
        producer.produce(
            topic=topic,
            value=self.serializer.encode(record),
            key=record.partition_key,
            on_delivery=ack
        )

        # wait callback and handle timeout exception
        try:
            future_result.result(timeout=_ms2sec_future(timeout_ms))
        except TimeoutError:
            raise MessagingError(f"Timeout waiting for message to be produced.")
        except Exception as e:
            raise MessagingError(f"Error producing message.") from e

    def write(self, record: Record, topic: str, timeout_ms=None) -> None:
        """
        Write a record to kafka using the transactional producer, according to current transactional state.
        """
        self._check_initialized()

        if self.in_transaction:
            self._write(self.transactional_producer, record, topic, timeout_ms)
        else:
            self._write(self.immediate_producer, record, topic, timeout_ms)

    def write_now(self, record: Record, topic: str, timeout_ms=None) -> None:
        """
        Write a record to kafka using the immediate producer, independently of the eventual current transaction.
        """
        self._check_initialized()

        self._write(self.immediate_producer, record, topic, timeout_ms)

    def start_transaction(self) -> None:
        """
        Enter in a transactional state (read/write).
        """
        self._check_initialized()

        if self.in_transaction:
            raise MessagingError(f"Already in a transaction.")
        self.transactional_producer.begin_transaction()
        self.in_transaction = True

    def commit(self, timeout_ms: Optional[int] = None) -> None:
        """
        Commit the current transaction.
        """
        self._check_initialized()

        if not self.in_transaction:
            raise MessagingError(f"Not in a transaction.")
        self.transactional_producer.commit_transaction(_ms2sec_ck(timeout_ms))
        logger.info(f"Committed transaction.")
        self.in_transaction = False

    def abort(self, timeout_ms: Optional[int] = None) -> None:
        """
        Abort the current transaction.

        Messages written in the transaction will be discarded. Offsets read in the transaction will be discarded.
        """
        self._check_initialized()

        if not self.in_transaction:
            raise MessagingError(f"Not in a transaction.")
        self.transactional_producer.abort_transaction(_ms2sec_ck(timeout_ms))
        logger.info(f"Aborted transaction.")
        self.in_transaction = False

    def __del__(self):
        if self.initialized:
            logger.info(f"Closing consumer.")
            try:
                self.consumer.close()
            except Exception:
                logger.warning(f"Error closing consumer.")
            logger.info(f"Waiting poll thread to join.")
            self.cancelled = True
            try:
                self.poll_thread.join()
            except Exception as e:
                logger.warning(f"Error joining poll thread: {e}.")


def _ms2sec_ck(timeout_ms: Optional[int]) -> float:
    """
    Convert milliseconds (integer) to seconds (float) for Confluent Kafka API.

    :param timeout_ms: timeout in milliseconds, None for no timeout
    :return: timeout in seconds, -1 for no timeout
    >>> _ms2sec_ck(None)
    -1.0
    >>> _ms2sec_ck(0)
    0.0
    >>> _ms2sec_ck(500)
    0.5
    """
    timeout_sec = -1.0
    if timeout_ms is not None:
        timeout_sec = timeout_ms / 1000
    return timeout_sec


def _ms2sec_future(timeout_ms: Optional[int]) -> Optional[float]:
    """
    Convert milliseconds (integer) to seconds (float) for the Future API.

    :param timeout_ms: timeout in milliseconds, None for no timeout
    :return: timeout in seconds, None for no timeout
    >>> _ms2sec_future(None)
    >>> _ms2sec_future(0)
    0.0
    >>> _ms2sec_future(500)
    0.5
    """
    timeout_sec = None
    if timeout_ms is not None:
        timeout_sec = timeout_ms / 1000
    return timeout_sec



# Copyright (c) Qotto, 2021

from __future__ import annotations

from datetime import datetime
from time import sleep
from typing import Iterable, Optional, Union

from eventy.messaging.errors import MessagingError
from eventy.messaging.store import RecordStore, Cursor
from eventy.record import Record

__all__ = [
    'MemoryStore',
]

from eventy.serialization import RecordSerializer
from eventy.serialization.dummy import DummySerializer


class _ReadTopic:
    """
    Topic that can be read from.

    Records are: [0..acked_offset..read_offset..len(records)]
    - 0..acked_offset: messages that have been acknowledged
    - acked_offset..read_offset: messages that have been read
    - read_offset..len(records): messages that have not been read yet
    """

    def __init__(
        self,
        topic: str,
        record_serializer: RecordSerializer,
    ) -> None:
        self.serializer = record_serializer
        self.topic = topic
        self.encoded_records: list[bytes] = []
        self.read_offset = 0
        self.acked_offset = 0

    def add(
        self,
        records: Union[Record, Iterable[Record]],
    ) -> None:
        if not isinstance(records, Iterable):
            records = [records]
        self.encoded_records.extend(map(self.serializer.encode, records))

    def read(
        self,
        max_count: int = 1,
    ) -> list[Record]:
        read_records_encoded = self.encoded_records[self.read_offset:self.read_offset + max_count]
        read_records = list(map(self.serializer.decode, read_records_encoded))
        self.read_offset += len(read_records)
        return read_records

    def ack(self, count: int = None) -> None:
        if count is None:
            self.acked_offset = self.read_offset
        else:
            if (self.acked_offset + count) > self.read_offset:
                raise MessagingError('Cannot ack more than read.')
            self.acked_offset += count

    def get_acked(self, pop: bool = False) -> list[bytes]:
        acked = self.encoded_records[:self.acked_offset]
        if pop:
            self.encoded_records = self.encoded_records[self.acked_offset:]
            self.read_offset -= len(acked)
            self.acked_offset -= len(acked)
        return list(acked)

    def get_read_unacked(self, pop: bool = False) -> list[bytes]:
        read_unacked = self.encoded_records[self.acked_offset:self.read_offset]
        if pop:
            self.encoded_records = list(
                self.encoded_records[0:self.acked_offset]
                + self.encoded_records[self.acked_offset:self.read_offset]
            )
            self.read_offset -= len(read_unacked)
        return list(read_unacked)

    def get_unread(self, pop: bool = False) -> list[bytes]:
        unread = self.encoded_records[self.read_offset:]
        if pop:
            self.encoded_records = self.encoded_records[0:self.read_offset]
        return list(unread)

    def remove_acked_records(self):
        self.encoded_records = self.encoded_records[self.acked_offset:]
        self.read_offset -= self.acked_offset
        self.acked_offset = 0

    def reset_to_acked(self):
        self.read_offset = self.acked_offset

    def clear(self) -> None:
        self.encoded_records = []
        self.read_offset = 0
        self.acked_offset = 0


class _WriteTopic:
    """
    Topic that can be written to.

    - records 0..committed_offset are committed.
    - records committed_offset..len(records) are not committed.
    """

    def __init__(
        self,
        topic: str,
        record_serializer: RecordSerializer,
    ) -> None:
        self.serializer = record_serializer
        self.topic = topic
        self.encoded_records: list[bytes] = []
        self.committed_offset = 0

    def write_committed(
        self,
        record: Record,
    ) -> None:
        self.encoded_records.insert(self.committed_offset, self.serializer.encode(record))
        self.committed_offset += 1

    def write_uncommitted(
        self,
        record: Record,
    ) -> None:
        self.encoded_records.append(self.serializer.encode(record))

    def commit(self, count: int = None) -> None:
        if count is None:
            self.committed_offset = len(self.encoded_records)
        else:
            if (self.committed_offset + count) > len(self.encoded_records):
                raise MessagingError("Cannot commit more than written.")
            self.committed_offset += count

    def get_committed(self, pop: bool = False) -> list[bytes]:
        committed = self.encoded_records[:self.committed_offset]
        if pop:
            self.encoded_records = self.encoded_records[self.committed_offset:]
            self.committed_offset -= len(committed)
        return list(committed)

    def get_uncommitted(self, pop: bool = False) -> list[bytes]:
        uncommitted = self.encoded_records[self.committed_offset:]
        if pop:
            self.encoded_records = self.encoded_records[0:self.committed_offset]
        return list(uncommitted)

    def reset_to_committed(self) -> None:
        self.encoded_records = self.encoded_records[:self.committed_offset]

    def remove_committed_records(self):
        self.encoded_records = self.encoded_records[self.committed_offset:]
        self.committed_offset = 0

    def clear(self) -> None:
        self.encoded_records = []
        self.committed_offset = 0


class _Transaction:
    def __init__(self):
        self.reads: dict[str, int] = {}
        self.writes: dict[str, int] = {}

    def add_read(self, topic: str, count: int = 1) -> None:
        if topic not in self.reads:
            self.reads[topic] = 0
        self.reads[topic] += count

    def add_write(self, topic: str, count: int = 1) -> None:
        if topic not in self.writes:
            self.writes[topic] = 0
        self.writes[topic] += count


class MemoryStore(RecordStore):
    """
    In-memory record store, essentially for testing and debug purposes
    """

    def __init__(
        self,
        record_serializer: RecordSerializer = None,
    ) -> None:
        """
        Initialize an empty store.

        :param record_serializer: Serializer to use for encoding and decoding records.
        """
        super().__init__()

        self.record_serializer = record_serializer or DummySerializer()
        self.read_topics: dict[str, _ReadTopic] = {}
        self.write_topics: dict[str, _WriteTopic] = {}
        self.transaction: Optional[_Transaction] = None

    def register_topic(self, topic: str, cursor: Cursor = Cursor.ACKNOWLEDGED) -> None:
        if cursor != Cursor.ACKNOWLEDGED:
            raise MessagingError(f"Only ACKNOWLEDGED cursors are supported.")
        super().register_topic(topic, cursor)
        self.read_topics[topic] = _ReadTopic(topic, record_serializer=self.record_serializer)

    def read(
        self,
        max_count: int = 1,
        timeout_ms: Optional[int] = None,
        auto_ack: bool = False
    ) -> list[Record]:
        if auto_ack:
            raise MessagingError(f"Auto-ack is not supported.")
        records: list[Record] = []
        for topic, topic_read in self.read_topics.items():
            reads = topic_read.read(max_count - len(records))
            records.extend(reads)
        return records

    def ack(self, timeout_ms=None) -> None:
        for topic, topic_read in self.read_topics.items():
            if self.transaction:
                self.transaction.add_read(topic, topic_read.read_offset - topic_read.acked_offset)
            else:
                topic_read.ack()

    def write(self, record: Record, topic: str, timeout_ms=None) -> None:
        if topic not in self.write_topics:
            self.write_topics[topic] = _WriteTopic(topic, record_serializer=self.record_serializer)
        write_topic = self.write_topics[topic]
        if self.transaction:
            self.transaction.add_write(topic)
            write_topic.write_uncommitted(record)
        else:
            write_topic.write_committed(record)

    def write_now(self, record: Record, topic: str, timeout_ms=None) -> None:
        if topic not in self.write_topics:
            self.write_topics[topic] = _WriteTopic(topic, record_serializer=self.record_serializer)
        write_topic = self.write_topics[topic]
        write_topic.write_committed(record)

    def start_transaction(self) -> None:
        if self.transaction:
            raise MessagingError(f"Already in a transaction.")
        self.transaction = _Transaction()

    def commit(self, timeout_ms: Optional[int] = None) -> None:
        if not self.transaction:
            raise MessagingError(f"Not in a transaction.")
        for topic, read_count in self.transaction.reads.items():
            self.read_topics[topic].ack(read_count)
        for topic, write_count in self.transaction.writes.items():
            self.write_topics[topic].commit(write_count)
        self.transaction = None

    def abort(self, timeout_ms: Optional[int] = None) -> None:
        if not self.transaction:
            raise MessagingError(f"Not in a transaction.")
        for topic, read_count in self.transaction.reads.items():
            self.read_topics[topic].reset_to_acked()
        for topic, write_count in self.transaction.writes.items():
            self.write_topics[topic].reset_to_committed()
        self.transaction = None

    def add_record(self, records: Union[Record, Iterable[Record]], topic: str) -> None:
        """
        Add one or multiple records to be read.

        :param records: record(s) to be added
        :param topic: topic the record(s) should be added to
        """
        if topic not in self.read_topics:
            raise MessagingError(f"Topic {topic} not registered.")
        self.read_topics[topic].add(records)

    def get_encoded(
        self,
        # options
        topic: str = None,
        pop: bool = False,
        # written
        committed: bool = False,
        uncommitted: bool = False,
        # read or available
        acked: bool = False,
        unacked: bool = False,
        unread: bool = False,
    ) -> list[bytes]:
        records: list = self.get(
            decode=False,
            topic=topic,
            pop=pop,
            committed=committed,
            uncommitted=uncommitted,
            acked=acked,
            unacked=unacked,
            unread=unread,
        )
        return records

    def get_records(
        self,
        # options
        topic: str = None,
        pop: bool = False,
        # written
        committed: bool = False,
        uncommitted: bool = False,
        # read or available
        acked: bool = False,
        unacked: bool = False,
        unread: bool = False,
    ) -> list[Record]:
        records: list = self.get(
            decode=True,
            topic=topic,
            pop=pop,
            committed=committed,
            uncommitted=uncommitted,
            acked=acked,
            unacked=unacked,
            unread=unread,
        )
        return records

    def get(
        self,
        # options
        topic: str = None,
        decode: bool = True,
        pop: bool = False,
        # written
        committed: bool = False,
        uncommitted: bool = False,
        # read or available
        acked: bool = False,
        unacked: bool = False,
        unread: bool = False,
    ) -> Union[list[Record], list[bytes]]:
        """
        Get one or multiple records from the store or a specific topic.

        :param topic: Topic, default: all topics
        :param decode: Decode (deserialize) the record(s)
        :param pop: Remove the record(s) from the store
        :param committed: Get the committed written records
        :param uncommitted: Get the uncommitted written records
        :param acked: Get the acknowledged read records
        :param unacked: Get the unacknowledged read records
        :param unread: Get the unread available records
        :return: A list of records or a list of encoded records
        """
        if topic:
            topics = [topic]
        else:
            topics = list(self.write_topics.keys()) + list(self.read_topics.keys())
        records: list = []

        for topic in topics:
            read_topic = self.read_topics.get(topic)
            if read_topic:
                if acked:
                    records.extend(read_topic.get_acked(pop=pop))
                if unacked:
                    records.extend(read_topic.get_read_unacked(pop=pop))
                if unread:
                    records.extend(read_topic.get_unread(pop=pop))
            write_topic = self.write_topics.get(topic)
            if write_topic:
                if committed:
                    records.extend(write_topic.get_committed(pop=pop))
                if uncommitted:
                    records.extend(write_topic.get_uncommitted(pop=pop))

        if decode:
            return list(map(self.record_serializer.decode, records))
        else:
            return list(records)

    def get_committed_records(self, topic: str = None, pop: bool = False) -> list[Record]:
        return self.get_records(topic=topic, pop=pop, committed=True)

    def get_all_written_records(self, topic: str = None, pop: bool = False) -> list[Record]:
        return self.get_records(topic=topic, pop=pop, committed=True, uncommitted=True)

    def get_acked_records(self, topic: str = None, pop: bool = False) -> list[Record]:
        return self.get_records(topic=topic, pop=pop, acked=True)

    def get_all_read_records(self, topic: str = None, pop: bool = False) -> list[Record]:
        return self.get_records(topic=topic, pop=pop, acked=True, unacked=True)

    def get_non_acked_records(self, topic: str = None, pop: bool = False) -> list[Record]:
        return self.get_records(topic=topic, pop=pop, unread=True, unacked=True)

    def wait_all_acked(self, timeout_ms: int = None) -> bool:
        """
        Wait until all records are acked.

        :param timeout_ms: Timeout in milliseconds
        :return: True if all records are acked, False if timeout
        """

        begin_time = datetime.now().timestamp()
        while self.get_non_acked_records():
            if timeout_ms and datetime.now().timestamp() > begin_time + timeout_ms / 1000.0:
                return False
            sleep(0.1)
        return True

    def clear(self) -> None:
        """
        Completely clear the store (remove all data).
        """
        for write_topic in self.write_topics.values():
            write_topic.clear()
        for read_topic in self.read_topics.values():
            read_topic.clear()
        self.transaction = None

    def reset(self) -> None:
        """
        Reset the store to the last committed state.
        """
        for write_topic in self.write_topics.values():
            write_topic.reset_to_committed()
        for read_topic in self.read_topics.values():
            read_topic.reset_to_acked()
        self.transaction = None

    def debug_topics(self, topic: str = None) -> str:
        result = ""
        if topic:
            topics = [topic]
        else:
            topics = list(self.read_topics.keys()) + list(self.write_topics.keys())
        max_len = max(map(len, topics))
        for topic in topics:
            if topic in self.write_topics:
                result += "READ  " + topic + (' ' * (max_len - len(topic))) + ": "
                result += ("X" * len(self.get(topic, acked=True)))
                result += ("0" * len(self.get(topic, unacked=True)))
                result += ("-" * len(self.get(topic, unread=True)))
                result += "\n"
            if topic in self.read_topics:
                result += "WRITE " + topic + (' ' * (max_len - len(topic))) + ": "
                result += ("X" * len(self.get(topic, committed=True)))
                result += ("O" * len(self.get(topic, uncommitted=True)))
                result += "\n"
        return result



# Copyright (c) Qotto, 2021

from __future__ import annotations

from typing import Optional

__all__ = [
    'Service',
]

from eventy.record import RecordType


class Service:
    def __init__(
        self,
        name: str,
        namespace: Optional[str] = None,
        event_topic: Optional[str] = None,
        request_topic: Optional[str] = None,
        response_topic: Optional[str] = None,
    ):
        """
        Initialize a service

        :param name: Name of the service
        :param namespace: Namespace URN of the service. Default: 'urn:<name>'
        :param event_topic: EVENT topic. Default: '<name>-events'
        :param request_topic: REQUEST topic. Default: '<name>-requests'
        :param response_topic: RESPONSE topic. Default: '<name>-responses'
        """
        self._name = name

        if namespace is None:
            namespace = f'urn:{name}'
        self._namespace = namespace

        if event_topic is None:
            event_topic = f'{name}-events'
        self._event_topic = event_topic

        if request_topic is None:
            request_topic = f'{name}-requests'
        self._request_topic = request_topic

        if response_topic is None:
            response_topic = f'{name}-responses'
        self._response_topic = response_topic

    @property
    def event_topic(self) -> str:
        return self._event_topic

    @property
    def request_topic(self) -> str:
        return self._request_topic

    @property
    def response_topic(self) -> str:
        return self._response_topic

    def topic_for(self, record_type: RecordType) -> str:
        if record_type == RecordType.EVENT:
            return self.event_topic
        elif record_type == RecordType.REQUEST:
            return self.request_topic
        elif record_type == RecordType.RESPONSE:
            return self.response_topic
        raise ValueError(f"No topic for record type {record_type}.")

    @property
    def namespace(self) -> str:
        return self._namespace

    @property
    def name(self) -> str:
        return self._name

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other) -> bool:
        return (
            other is not None
            and isinstance(other, Service)
            and self.name == other.name
            and self.namespace == other.namespace
            and self.request_topic == other.request_topic
            and self.response_topic == other.response_topic
            and self.event_topic == other.event_topic
        )

    def __str__(self) -> str:
        return f'<Service {self.name} ({self.namespace})>'



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['eventy',
 'eventy.config',
 'eventy.config.aiohttp',
 'eventy.config.celery',
 'eventy.config.django',
 'eventy.config.sanic',
 'eventy.integration',
 'eventy.logging',
 'eventy.messaging',
 'eventy.record',
 'eventy.serialization',
 'eventy.trace_id']

package_data = \
{'': ['*']}

install_requires = \
['coloredlogs>=15', 'contextvars>=2', 'semver>=2']

extras_require = \
{'aiohttp': ['aiohttp>=1'],
 'avro': ['PyYAML>=5', 'avro-python3>=1.10'],
 'celery': ['celery>=5'],
 'confluent-kafka': ['confluent-kafka>=1.7'],
 'django': ['django>=1'],
 'requests': ['requests>=2'],
 'sanic': ['sanic>=18']}

setup_kwargs = {
    'name': 'eventy',
    'version': '3.2.1',
    'description': 'Qotto/eventy',
    'long_description': "# Eventy\n\n* Source: [Eventy on GitLab](https://gitlab.com/qotto/oss/eventy)\n* Package: [Eventy on Pypi](https://pypi.org/project/eventy/)\n* Documentation: [Eventy API documentation](https://qotto.gitlab.io/oss/eventy/)\n\n## What is Eventy?\n\nEventy is both a protocol and a library for making the design of fault-tolerant, event-driven, concurrent and\ndistributed applications in a microservices-oriented architecture easier.\n\nAs a protocol, Eventy is language-agnostic.\n\nHowever, the reference implementation is written in Python. Whatever is your programming language, you can use Eventy.\nIf you are using Python, you will be able to make the most of it quicker.\n\n## Motivation\n\nThe reality is, a distributed, microservices-oriented architecture is a bit hard to make right.\n\nMartin Fowler even [ended up stating](https://www.drdobbs.com/errant-architectures/184414966):\n\n> First Law of Distributed Object Design: “don't distribute your objects”.\n\nHe later [detailed his view](https://martinfowler.com/articles/distributed-objects-microservices.html) of the First Law\nregarding microservices.\n\nAs Martin Folwer points out, inherently microservices come with their undeniable advantages and a few companies use them\nwith great success. The only reason why a lot of people end up struggling with microservices, is because it greatly\nincreases the complexity of the whole system, which makes it harder to get it right.\n\nEventy is adressing exactly that issue — providing you with a reliable framework to make microservices work for you.\n\nA big part of the design process to create Eventy was to understand different use cases and see how the design can be\nsimplified. For instance, not only does Eventy offer state partitionning for free, it actually comes up with a few\nstratagies that eliminate the need to persist state altogether.\n\n## Inspiration\n\n[Kafka Streams](https://kafka.apache.org/documentation/streams/) was a great influence in making\nEventy. [Celery](http://www.celeryproject.org/) and [Faust](https://github.com/robinhood/faust) are also worth to be\nlooked at if you are looking for an opiniated framework easy to get started with.\n\nHowever, these frameworks only partially solve all the issues you will have with microservices. And, in our opinion,\nthese frameworks are not suitable for designing large critical systems.\n\nThey're both opinanated, and therefore cannot be easily integrated in your existing software. You will have to build\nyour software around the framework, instead of the other way around. They also don't give you the full control on the\nway you can use them: you can only use them as a whole, or not at all.\n\n## What Eventy can do for you\n\nEventy implements multiple features, but all of them simply solve two main problems:\n\n* How to make services communicate with each other\n* How to access and persist state\n\nWith Eventy, you can serialize data the way you want. You can use [Avro](https://avro.apache.org/)\n, [JSON](https://www.json.org/), [gRPC](https://grpc.io/), or whatever customer serializer you like.\n\nWith Eventy, you can use any system you like as a persistency layer, as long as it supports transactions, if you need\nstrong processing guarantees. The most obvious choice is to use [Apache Kafka](https://kafka.apache.org/), but\npersisting messages over [PostgreSQL](https://www.postgresql.org/) is completely feasable, too.\n\nEventy was destined with the mindset of a library of related but independently usable components - and not a framework:\nthe behaviour is explicit and you're the one in charge: you can design your software your own way.\n\nThis explicit behaviour, albeit requiring more boilerplate, gives you better clarity on what is happening. Recipes and\nexamples are provided so that you can understand how to use Eventy for most use cases.\n\nYou're free to use any part of Eventy as well. Even if you end up not using the Eventy protocol at all, simply reading\nthe documentation and understanding the issues that are adressed and how they are adressed can help you to get on the\nright path.\n\n## Main components of Eventy\n\n* a **well-defined communication protocol** for sending various types of persisted messages, called _Records_: _Events_\n  , _Requests_ and _Responses_\n* **persistency of _Records_** that can be stored forever, which lets you keep track of all the changes in your system (\n  especially useful for audits and business analytics)\n* **queues** so that _Records_ can be processed asynchroneously, and aren't lost even if your system is down or\n  overloaded\n* **strong processing guarantees**: a Record can be designed to be processed _at least once_, _at most once_ and _\n  exactly once_ even if your system encounters a process or network failure at any point\n* **self-propagating _Contexts_** that in many cases entirely eliminate the need of persisting state\n* **partitionned state persistency** so that you no longer have a single point of failure in your system (aka _the\n  database_) and can scale up as your business grows\n\n# Contribute\n\nInstall with dev dependencies\n\n```\npoetry install -E celery -E sanic -E aiohttp -E django -E confluent-kafka -E avro -E requests\n```\n\nThe project uses [poetry](https://python-poetry.org/)\n\n```\npoetry config pypi-token.pypi my-token\npoetry config testpypi-token.pypi my-token\n```\n",
    'author': 'Qotto Dev Team',
    'author_email': 'dev@qotto.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/qotto/oss/eventy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
